--[[
---------------------------------------------------------------------------
Remote Train Control Program for Windows

© Copyright 2024 by Mark DiVecchio

	This file is part of Remote Train Control.

	Remote Train Control is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Remote Train Control is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with Remote Train Control.  If not, see <http://www.gnu.org/licenses/>.

 Home Page : http://www.silogic.com/trains/RTC_Running.html
---------------------------------------------------------------------------]]
-- Semicolons are not required in Lua code
title = "Layout4 APB Master"
require([[defines]]);
require([[functions]]);

--[[
When using this module, all variables (including functions!) must be declared through 
a regular assignment (even assigning nil will do) in a strict scope before being used 
anywhere or assigned to inside a nested scope.
https://github.com/lua-stdlib/strict
]]
local _ENV = require 'std.strict' (_G)	-- strict.lua will detect undeclared global variables
local _;
--
--[[
---------------------------------------------------------------------------
Based Program for APB example defined in August 2016 RMC by Bruce Chubb
Version Revision 07/17/2016

Initialize signals to red to cover possibile single track
occupancies at startup resulting in false setting of
traffic sticks, i.e. sticks set for both directions
when shuld be no traffic sticks set

The APB system does not have separate occupancy information for passing
sidings. If either the thru or passing track is occupied, the information
is stored with the thru track.

Any reference to a passing/station track is translated to the thru track
for use by this script.
---------------------------------------------------------------------------]]

--[[
For testing, accepts events from "Layout4 Block XX Occupancy.lua" to simulate a train running across the layout
--]]

local Count = 0;

local Tags = {};		-- table to hold all of the tags detected in sequence
local TagCount = 0;

local FSR = {};			-- table to hold all of the traffic sticks
local Sig = {};			-- table to hold all of the Next Signal Aspects
local SavedSig = {};	-- table to hold the last known Signal Aspects
local BlockOccupancy = {};			-- Table to hold the APB Master Block Occupancy Status, OCCUPIED or UNOCCUPIED
									-- Values in this table can be:
									-- -2	SOFTOCCUPY
									-- -1	MANUALFLAG
									--  0	UNOCCUPIED
									-- >0	OCCUPIED
									--		1	Occupied or a thru/passing with 1 branch occupied
									--		2	Thru/passing with both branches occupied
									
local TUMCNT = {};		-- table to hold all of the Tumbledown Counters
--
-- Traffic Stick Directions
local WEST = 1;
local EAST = 2;
local NDT = 0;		-- No direction-of-traffic
-- state machine states
local IDLE = 0;
local SM = IDLE;
local ErrCount = 0;
local BeepOn = false;
local SoundOff = false;
--
-- Instrumented Layout 4 in global variable
Layout4 = nil;
-- clear all tag handling functions
local Func30 = nil;
local Func51 = nil;
local Func10 = nil;
local Func41 = nil;
local Func21 = nil;
local Func20 = nil;
local Func90 = nil;
local Func40 = nil;
local Func50 = nil;
local Func60 = nil;
local Func61 = nil;
local Func100 = nil;
local Func101 = nil;
--
local MyTIUNo, Enum;
local OccupiedBlocks = {};		-- list of occupied blocks read from settings file
--
local SetupComplete = false;	-- if true, permits settings to be saved.

-- These labels appear on each of the tabs for the Function Buttons
Tab01Label = "Control";
Tab02Label = "Support";
Tab03Label = "Debug";
--[[---------------------------------------------------------------------------------------]]
-- settings are saved in this file
local settingsfile = "APBMaster-PC" .. tostring(PCNo()) .. ".ini";

local function readSettings()
	OccupiedBlocks = {};		-- clear the occupied blocks list
    local f = io.open(settingsfile, "r");
    if f then
		print("Loading settings from " .. settingsfile);
		-- read occupied blocks from settings file if the file exists
		for iBlk = 1, BLOCK_COUNT do	-- up to BLOCK_COUNT blocks
			local BlockNo = f:read("*l");
			if (BlockNo == nil) then
				break;
				end;
			local iBlockNo = tonumber(string.match(BlockNo, "%d+"));
			if (iBlockNo == nil) then
				break;
				end;
			if (iBlockNo <= LAST_BLOCK) then
				print("Previously Occupied Block = " .. iBlockNo .. " " .. BlockName[iBlockNo]);
				OccupiedBlocks[iBlk] = iBlockNo;
				end;
			end;
        f:close()
		return true;
    end
	return false;
end

--[[---------------------------------------------------------------------------------------]]

local function writeSettings(pflag)
	-- setup must be complete before settings can be saved
	-- This function will also save manually set Block Occupancy flags - you probably don't want that
	-- pflag == false for less printed output
	if (not SetupComplete) then
		return false;
		end;
    local f = io.open(settingsfile, "w");
    if f then
		print("Saving settings in " .. settingsfile);
		-- write occupied blocks to settings file
		for iBlk = 1, BLOCK_COUNT do	-- up to BLOCK_COUNT blocks
			if ((BlockOccupancy[iBlk] ~= nil) and (BlockOccupancy[iBlk] > 0)) then
				for iflagcount = 1, BlockOccupancy[iBlk] do
					-- Passing/station blocks can have a BlockOccupancy[] count of 0, 1, or 2,
					-- which, if 2, can generate multiple lines in the settings file.
					-- Blocks whose flags are MANUALFLAG or SOFTOCCUPY are not written
					if (pflag) then print("Currently Occupied Block = " .. iBlk .. " " .. BlockName[iBlk]); end;
					f:write(tostring(iBlk) .. " " .. BlockName[iBlk] .. "\n");
					end;
				end;
			end;
        f:close()
		return true;
    end
	return false;
end
--[[---------------------------------------------------------------------------------------]]
--[[---------------------------------------------------------------------------------------]]
function setup(Engine, TIU)
	print("setup() Running");
	-- Instrumented Layout 4
	Layout4 = require([[Layout4]]);		-- All of the details about the Layout
	--
	-- APB Master will work with either the Looping or the Point-to-Point Layout Configuration
	--	
	-- APB Master does not distinguish between "thru" and "passing/station" tracks they
	-- are considered as one block
	-- fix up passing/station blocks names
	BlockName[STRUTHERS_THRU_TRACK] = "Struthers Block";
	BlockName[STRUTHERS_STATION_TRACK] = "Struthers Block";
	BlockName[COLLEGE_THRU_TRACK] = "College Block";
	BlockName[COLLEGE_STATION_TRACK] = "College Block";
	BlockName[J_L_PASSING_TRACK] = "J&L Block";
	BlockName[J_L_THRU_TRACK] = "J&L Block";
	BlockName[GATEWAY_YARD1_TRACK] = "Gateway Yard Block";
	BlockName[GATEWAY_YARD2_TRACK] = "Gateway Yard Block";
	BlockName[MCKEES_ROCKS_REPAIR_SHOPS_YARD1_TRACK] = "McKees Rocks Repair Shops Block";
	BlockName[MCKEES_ROCKS_REPAIR_SHOPS_YARD2_TRACK] = "McKees Rocks Repair Shops Block";
	-- Possible better way to write it
	--BlockName[MCKEES_ROCKS_REPAIR_SHOPS_YARD2_TRACK] = string.gsub(BlockName[MCKEES_ROCKS_REPAIR_SHOPS_YARD2_TRACK], " 2 ", " ");
	
	
	-- The TIU which controls layout 4 is defined in the Layout4.lua file	
	MyTIUNo = Layout4.MASTER_TIU;
	-- read in the occupied blocks from previous run
	readSettings();
	--
	-- Requires 8 detectors and 13 tag readers
	local NumDetectors, NumReaders;
	NumDetectors, NumReaders = GetTagReaders();
	print("Number of Tag Detectors = " .. NumDetectors .. " Number of Tag Readers =  " .. NumReaders);
	if ((Debug() == 0) and (NumDetectors ~= 8 or (NumReaders ~= 13 and NumReaders ~= 0))) then
		return Stop("Requires 8 Tag Detectors and 13 Tag Readers. Cycle power on layout and try again.");
		end;
	--	
	print("Showing the Layout Window");
	ShowLayout();		-- show the Layout Window
	--
	-- Next line not needed for normal operation - This script should run independently from Engine control scripts
	-- It is used for testing with "Layout4 APB XX Occupancy.lua" scripts only
	ReqBlockEvents(true);	-- request block occupancy flag events from the Layout to keep our data up to date	
	--
	DeleteAllSignals();	-- clear the window of any old signals
	print("Creating Signals");
	-- Signals in each direction
	-- Most signals are Permissive
	-- Signals controlling entrance into head blocks from sidings are Absolute
	-- Route runs EAST <-> WEST
-----------------------------------------------------------------------------------------------
--[[	
	Initialize signals to red to cover possibile single track
    occupancies at startup resulting in false setting of
    traffic sticks, i.e. sticks set for both directions
    when there should be no traffic sticks set
]]   
-----------------------------------------------------------------------------------------------
--[[
	Remember that switches set to REVERSE appear as an occupied block
	It may be nessary to set switches to NORMAL to allow trains
	to move into that block.
]]
-----------------------------------------------------------------------------------------------
-- Set Loop Switch
	if (Layout4.LLLoop == true) then
		-- Layout using the bypass block - Balloon Block/Lionel-Ives Block - if Layout4.LLLoop = true
		print("Loop Configuration");		-- Loop Configuration
		Switch(REVERSE, Switches[BALLOON][CHAN_NUMBER], Switches[BALLOON][AIU_NUMBER], Switches[BALLOON][TIU_NUMBER], 0)
		else
		-- Layout not using the bypass block - if Layout4.LLLoop = false
		print("Point-to-Point Configuration");	-- Point-to-Point Configuration
		Switch(NORMAL, Switches[BALLOON][CHAN_NUMBER], Switches[BALLOON][AIU_NUMBER], Switches[BALLOON][TIU_NUMBER], 0)
		end;
	-- Double toggle all passing/station track switch to loosen them op
	--print("All passing/station track switches to NORMAL");
	Switch(NORMAL, Switches[ALIQUIPPA_WEST][CHAN_NUMBER], Switches[ALIQUIPPA_WEST][AIU_NUMBER], Switches[ALIQUIPPA_WEST][TIU_NUMBER], 0)
	--print("Aliquippa West Switch to NORMAL"); Sleep(2);
	Switch(NORMAL, Switches[ALIQUIPPA_EAST][CHAN_NUMBER], Switches[ALIQUIPPA_EAST][AIU_NUMBER], Switches[ALIQUIPPA_EAST][TIU_NUMBER], 0)
	--print("Aliquippa East Switch to NORMAL"); Sleep(2);
	Switch(NORMAL, Switches[COLLEGE_WEST][CHAN_NUMBER], Switches[COLLEGE_WEST][AIU_NUMBER], Switches[COLLEGE_WEST][TIU_NUMBER], 0)
	--print("College West Switch to NORMAL"); Sleep(2);
	Switch(NORMAL, Switches[COLLEGE_EAST][CHAN_NUMBER], Switches[COLLEGE_EAST][AIU_NUMBER], Switches[COLLEGE_EAST][TIU_NUMBER], 0)
	--print("College East Switch to NORMAL"); Sleep(2);
	Switch(NORMAL, Switches[STRUTHERS_WEST][CHAN_NUMBER], Switches[STRUTHERS_WEST][AIU_NUMBER], Switches[STRUTHERS_WEST][TIU_NUMBER], 0)
	--print("Struthers West Switch to NORMAL"); Sleep(2);
	Switch(NORMAL, Switches[STRUTHERS_EAST][CHAN_NUMBER], Switches[STRUTHERS_EAST][AIU_NUMBER], Switches[STRUTHERS_EAST][TIU_NUMBER], 0)
	--print("Struthers East Switch to NORMAL"); Sleep(2);
	Sleep(3);	-- time for all of the switches to align
	
--	
-- Westbound
--
	CreateSignal( 1, WB, PERMISSIVE_IN_OPEN_COUNTRY, 370,280, SignalNames[1], AsStopandProceed, 2, 1);		-- Permissive signal in open country
	FSR[1] = NDT;
	Sig[1] = AsStopandProceed;
	TUMCNT[1] = 0;
--	
	CreateSignal( 5, WB, ABSOLUTE_HEAD_BLOCK, 800,260, SignalNames[5], AsStopSignal, 1, 0);		-- Absolute Head Block Signal
    FSR[5] = nil;
	Sig[5] = AsStopSignal;
	TUMCNT[5] = 0;
--
	CreateSignal( 9, WB, PERMISSIVE_HEAD_BLOCK, 1190,240, SignalNames[9], AsStopandProceed, 3, 9);	-- Permissive head block signal
	FSR[9] = NDT;
	Sig[9] = AsStopandProceed;
	TUMCNT[9] = 0;
--
	CreateSignal(11, WB, APPROACH_TO_HEAD_BLOCK, 1270,240, SignalNames[11], AsStopandProceed, 2, 11);	-- Signal in approach to head block signal
	FSR[11] = NDT;
	Sig[11] = AsStopandProceed;
	TUMCNT[11] = 0;
--	
	CreateSignal(13, WB, ABSOLUTE_HEAD_BLOCK, 1450,240, SignalNames[13], AsStopSignal, 1, 0);		-- Absolute Head Block Signal
	FSR[13] = NDT;
	Sig[13] = AsStopSignal;
	TUMCNT[13] = 0;
--
	CreateSignal(17, WB, PERMISSIVE_HEAD_BLOCK, 1750,240, SignalNames[17], AsStopandProceed, 3, 17);	-- Permissive head block signal
	FSR[17] = NDT;
	Sig[17] = AsStopandProceed;
	TUMCNT[17] = 0;
--
	CreateSignal(19, WB, APPROACH_TO_HEAD_BLOCK, 1833,240, SignalNames[19], AsStopandProceed, 2, 19);	-- Signal in approach to head block signal
	FSR[19] = NDT;
	Sig[19] = AsStopandProceed;
	TUMCNT[21] = 0;
--
	CreateSignal(21, WB, PERMISSIVE_IN_OPEN_COUNTRY, 2000,240, SignalNames[21], AsStopandProceed, 2, 21);	-- Permissive signal in open country
	FSR[21] = NDT;
	Sig[21] = AsStopandProceed;
	TUMCNT[21] = 0;
--
	CreateSignal(25, WB, PERMISSIVE_IN_OPEN_COUNTRY, 2100,240, SignalNames[25], AsStopandProceed, 2, 25);	-- Permissive signal in open country
	FSR[25] = NDT;
	Sig[25] = AsStopandProceed;
	TUMCNT[25] = 0;
--
	CreateSignal(27, WB, PERMISSIVE_IN_OPEN_COUNTRY, 2320,240, SignalNames[27], AsStopandProceed, 2, 27);	-- Permissive signal in open country
	FSR[27] = NDT;
	Sig[27] = AsStopandProceed;
	TUMCNT[27] = 0;
--
	CreateSignal(29, WB, ABSOLUTE_HEAD_BLOCK, 2675,240, SignalNames[29], AsStopSignal, 1, 0);		-- Absolute Head Block Signal
	FSR[29] = NDT;
	Sig[29] = AsStopSignal;
	TUMCNT[29] = 0;
--
	CreateSignal(33, WB, PERMISSIVE_HEAD_BLOCK, 3280,240, SignalNames[33], AsStopandProceed, 3, 33);	-- Permissive head block signal
	FSR[33] = NDT;
	Sig[33] = AsStopandProceed;
	TUMCNT[33] = 0;
--
	CreateSignal(37, WB, APPROACH_TO_HEAD_BLOCK, 3430,240, SignalNames[37], AsStopandProceed, 2, 37);	-- Signal in approach to head block signal
	FSR[37] = NDT;
	Sig[37] = AsStopandProceed;
	TUMCNT[37] = 0;
--
-----------------------------------------------------------------------------------------------
--
-- Eastbound
--	
	CreateSignal(2, EB, APPROACH_TO_HEAD_BLOCK, 370,440, SignalNames[2], AsStopandProceed, 2, 2);		-- Signal in approach to head block signal
	FSR[2] = NDT;
	Sig[2] = AsStopandProceed;
	TUMCNT[2] = 0;
--
	CreateSignal(6, EB, PERMISSIVE_HEAD_BLOCK, 800,430, SignalNames[6], AsStopandProceed, 3, 6);		-- Permissive head block signal
	FSR[6] = NDT;
	Sig[6] = AsStopandProceed;
	TUMCNT[6] = 0;
--
	CreateSignal(10, EB, ABSOLUTE_HEAD_BLOCK, 1190,410, SignalNames[10], AsStopSignal, 1, 0);		-- Absolute Head Block Signal
	FSR[10] = NDT;
	Sig[10] = AsStopSignal;
	TUMCNT[10] = 0;
--	
	CreateSignal(12, EB, APPROACH_TO_HEAD_BLOCK, 1270,410, SignalNames[12], AsStopandProceed, 2, 12);	-- Signal in approach to head block signal
	FSR[12] = NDT;
	Sig[12] = AsStopandProceed;
	TUMCNT[12] = 0;	
--
	CreateSignal(14, EB, PERMISSIVE_HEAD_BLOCK, 1450,410, SignalNames[14], AsStopandProceed, 3, 14);	-- Permissive head block signal
	FSR[14] = NDT;
	Sig[14] = AsStopandProceed;
	TUMCNT[14] = 0;
--
	CreateSignal(18, EB, ABSOLUTE_HEAD_BLOCK, 1750,410, SignalNames[18], AsStopSignal, 1, 0);		-- Absolute Head Block Signal
	FSR[18] = NDT;
	Sig[18] = AsStopSignal;
	TUMCNT[18] = 0;
--
	CreateSignal(20, EB, PERMISSIVE_IN_OPEN_COUNTRY, 1833,410, SignalNames[20], AsStopandProceed, 2, 20);	-- Permissive signal in open country
	FSR[20] = NDT;
	Sig[20] = AsStopandProceed;
	TUMCNT[20] = 0;
--
	CreateSignal(22, EB, PERMISSIVE_IN_OPEN_COUNTRY, 2000,410, SignalNames[22], AsStopandProceed, 2, 22);	-- Permissive signal in open country
	FSR[22] = NDT;
	Sig[22] = AsStopandProceed;
	TUMCNT[22] = 0;
--
	CreateSignal(26, EB, PERMISSIVE_IN_OPEN_COUNTRY, 2100,410, SignalNames[26], AsStopandProceed, 2, 26);	-- Permissive signal in open country
	FSR[26] = NDT;
	Sig[26] = AsStopandProceed;
	TUMCNT[26] = 0;
--
	CreateSignal(28, EB, APPROACH_TO_HEAD_BLOCK, 2320,410, SignalNames[28], AsStopandProceed, 2, 28);	-- Signal in approach to head block signal
	FSR[28] = NDT;
	Sig[28] = AsStopandProceed;
	TUMCNT[28] = 0;
--	
	CreateSignal(30, EB, PERMISSIVE_HEAD_BLOCK, 2675,410, SignalNames[30], AsStopandProceed, 3, 30);	-- Permissive head block signal
	FSR[30] = NDT;
	Sig[30] = AsStopandProceed;
	TUMCNT[30] = 0;
--
	CreateSignal(34, EB, ABSOLUTE_HEAD_BLOCK, 3280,410, SignalNames[34], AsStopSignal, 1, 0);		-- Absolute Head Block Signal
	FSR[34] = NDT;
	Sig[34] = AsStopSignal;
	TUMCNT[34] = 0;
--
	CreateSignal(38, EB, PERMISSIVE_IN_OPEN_COUNTRY, 3430,410, SignalNames[38], AsStopandProceed, 2, 38);	-- Permissive signal in open country
	FSR[38] = NDT;
	Sig[38] = AsStopandProceed;
	TUMCNT[38] = 0;
--
	print("Signals Created");

-- save the current values
local orig_key, orig_value;
for orig_key, orig_value in pairs(Sig) do
	--print("Signal copy [", orig_key, "] = " , AspectNames[orig_value]);
    SavedSig[orig_key] = orig_value;	
	end;

-- clear all tag handling functions
Func30 = nil;
Func51 = nil;
Func10 = nil;
Func41 = nil;
Func21 = nil;
Func20 = nil;
Func90 = nil;
Func40 = nil;
Func50 = nil;
Func60 = nil;
Func61 = nil;
Func100 = nil;
Func101 = nil;

-- Unoccupy all Blocks
print("Set all Blocks as Unoccupied");
for iBlock = FIRST_BLOCK, LAST_BLOCK do
	BlockOccupancy[iBlock] = UNOCCUPIED;					-- this flag is used by this APB Master thread
	-- this script clears the global occupancy flags as we startup a run
	OverrideOccupancy(iBlock, UNOCCUPIED);	-- force the flag to show unoccupied 	-- This flag is used by the Engine PC threads
	::nextBlockUnoccupied::
	end;

--[[--------------------------------------------]]
--
-- Ask the dispatcher for occupied block number(s)
--
-- Note: when running with "Layout4 Block xB Occupancy.lua", it is not necessary to to select a starting block
-- as those test scripts will set the starting block number to occupied
--
local BlockNameList = {};	-- table to hold all of the block numbers and names
	
--BlockNameList[1] = "0 None";	-- use index 1 for "None"
BlockNameList[1] = NO_BLOCK .. " " .. BlockName[NO_BLOCK];	-- use index 1 for "None"
local iBlk = 2;
for iBlock = 1, BLOCK_COUNT do
	if (iBlock <= LAST_BLOCK) then
		-- in Point-to-Point mode, don't allow selection of Balloon Track
		if (Layout4.LLLoop == false and iBlock == BALLOON_BLOCK) then
			goto nextBlock
			end;
		-- in Loop mode, don't allow selection of the yard tracks
		if (Layout4.LLLoop == true) then
			if ((iBlock == GATEWAY_YARD_BLOCK) or (iBlock == MCKEES_ROCKS_REPAIR_SHOPS_YARD_BLOCK)) then
				goto nextBlock;
				end;
			end			
		if ((iBlock == J_L_PASSING_TRACK) or (iBlock == COLLEGE_STATION_TRACK) or (iBlock == STRUTHERS_STATION_TRACK) or
			(iBlock == GATEWAY_YARD2_TRACK) or (iBlock == MCKEES_ROCKS_REPAIR_SHOPS_YARD2_TRACK)) then		
			-- APB system treats THRU and STATION/PASSING and YARDx as the same track
			goto nextBlock;
			end;			
		--
		--print(Name .. "(" .. iBlock .. ")");
		BlockNameList[iBlk] = iBlock .. " " .. BlockName[iBlock];
		iBlk = iBlk + 1;
		::nextBlock::
		end;
	end;
--
--print("BlockNameList -------------------------------");
--PrettyPrint(BlockNameList);
--print("BlockNameList -------------------------------");
--
-- read Occupied Block Number(s)
print("Select Occupied Block Number");
--
-- read up to BLOCK_COUNT occupied blocks
--
for iBlk = 1, BLOCK_COUNT do
	local OccupiedBlockNo;
	-- read next block - note that there may be duplicates in this list in the case of passing sidings
	OccupiedBlockNo = OccupiedBlocks[iBlk];
	-- if no more blocks, use "None" index setting of 1
	if (OccupiedBlockNo == nil) then
		OccupiedBlockNo	= 0;
		end;
	local BlockNameIndex = 1;
	local BlockNameCount = 1;	
	
	for iBlock = 1, BLOCK_COUNT do
		local Name = BlockNameList[iBlock];
		if (Name ~= nil and iBlock <= LAST_BLOCK) then
			--print(Name .. " --> (" .. iBlock .. ")");
			--------------------------------------------
			local BlockNo = tonumber(string.match(Name, "%d+"));
			if (OccupiedBlockNo == BlockNo) then
				BlockNameIndex = BlockNameCount;
				--print("BlockNameIndex = " .. BlockNameIndex);
				end;
			BlockNameCount = BlockNameCount + 1;		
			--------------------------------------------		
			end;
		end;

	local OccupiedBlock = InputDropDown(title, "Select Occupied Block ('None' for no more)", BlockNameIndex, BlockNameList);
	--print("OccupiedBlock ", OccupiedBlock);
	local typ = type(OccupiedBlock);
	--print("Type ", typ);

	if (typ == "boolean" and OccupiedBlock == false) then
		return Stop("InputDropDown Canceled");
		else
		if (OccupiedBlock ~= nil and OccupiedBlock ~= "") then	-- if the result is a string, then use first number in the string, otherwise, ignore it
			-- pull first integer found in the string
			local BlockNo = string.match(OccupiedBlock, "%d+");
			if (BlockNo ~= nil) then
				OccupiedBlock = tonumber(BlockNo);
				if (OccupiedBlock == 0) then
					break;	-- done
					end;
				-- fixup for passing/station/yard tracks - algorithm considers the passing/station/yard tracks
				-- as occupied together with the thru tracks
				if (OccupiedBlock == STRUTHERS_STATION_TRACK) then OccupiedBlock = STRUTHERS_BLOCK; end;
				if (OccupiedBlock == COLLEGE_STATION_TRACK) then OccupiedBlock = COLLEGE_BLOCK; end;
				if (OccupiedBlock == J_L_PASSING_TRACK) then OccupiedBlock = J_L_BLOCK; end;
				if (OccupiedBlock == GATEWAY_YARD2_TRACK) then OccupiedBlock = GATEWAY_YARD_BLOCK; end;				
				if (OccupiedBlock == MCKEES_ROCKS_REPAIR_SHOPS_YARD2_TRACK) then OccupiedBlock = MCKEES_ROCKS_REPAIR_SHOPS_YARD_BLOCK; end;				
				-- indicate that the block is occupied.
				BlockOccupancy[OccupiedBlock] = BlockOccupancy[OccupiedBlock] + 1;	-- this flag is used by this APB Master thread			
				-- force the flag to show soft occupancy 
				OverrideOccupancy(OccupiedBlock, SOFTOCCUPY);						-- This flag is used by the EOB PC threads

				else
				return Stop("Invalid Block Number");
				end;
			else
			if (typ == "nil") then
				-- nil returned only if parameters are invalid
				return Stop("Invalid Parameters to InputDropDown");
				end
			end
		end
	print("Occupied Block = " .. BlockName[OccupiedBlock] .. " (" .. OccupiedBlock .. ")");
	end;
-- indicate that setup is complete and the settings can be saved
SetupComplete = true;
--
-- Stop the Garbage collector since this is a real-time program
collectgarbage("stop")
print("Setup : Garbage Collection stopped");
--collectgarbage("collect")
--collectgarbage("restart")
-- now as soon as the first engine passes over a reader, it will change the BlockOccupancy[] Occupancy flags and cause signals to change
	
return true;		-- false=setup failed, true=setup succeeded
end
--[[---------------------------------------------------------------------------------------]]
-- SET TRAFFIC STICK UPON SEQUENTIAL TRAIN MOVEMENT	
-- use (western side block, eastern side block, westbound signal #, eastbound signal #)
function SetTrafficStick(Block1, Block2, Sig1, Sig2)
if (((BlockOccupancy[Block1] ~= UNOCCUPIED) and (BlockOccupancy[Block2] ~= UNOCCUPIED)) and ((FSR[Sig1] == NDT) and (FSR[Sig2] == NDT))) then
	if (Sig[Sig1] ~= AsStopSignal and Sig[Sig1] ~= AsStopandProceed) then
		print("Traffic Stick at signal #", Sig1, " set to WEST");
		FSR[Sig1] = WEST;
		end;
	if (Sig[Sig2] ~= AsStopSignal and Sig[Sig2] ~= AsStopandProceed) then
		print("Traffic Stick at signal #", Sig2, " set to EAST");
		FSR[Sig2] = EAST;
		end;
	end;	
end;
--
local TUMMAX = 15;  --maximum value for tumbledown delay counter
--
--[=[
APB_Program_for_August_Part_8_RMC

Rem**CALCULATE EASTBOUND SIGNAL ASPECTS**
.
.
.
.
Rem**Calculate Signal 124 Permissive in open country
S124: If FS123 = WEST Then GoTo S124RED
      If BK15 = OCC Then GoTo S124RED
      If SIG136 <> RED Then SIG124 = GRN: TUMCNT124 = 0: GoTo S114
      If FS136 = EAST Then SIG124 = YEL: TUMCNT124 = 0: GoTo S114
S124RED: If SIG124 = RED Then GoTo S114
      If TUMCNT124 = TUMMAX Then SIG124 = RED: TUMCNT124 = 0: GoTo S142
      TUMCNT124 = TUMCNT124 + 1

Rem**Calculate Signal 114 Permissive in open country
S114: If FS113 = WEST Then GoTo S114RED
      If BK14 = OCC Then GoTo S114RED
      If SIG124 <> RED Then SIG114 = GRN: TUMCNT114 = 0: GoTo S106
      If FS124 = EAST Then SIG114 = YEL: TUMCNT114 = 0: GoTo S106
S114RED: If SIG114 = RED Then GoTo S106
      If TUMCNT114 = TUMMAX Then SIG114 = RED: TUMCNT114 = 0: GoTo S106
      TUMCNT114 = TUMCNT114 + 1

Rem**Calculate Signal 106 Head Block Absolute
S106: If BK13 = OCC Then GoTo S106RED
      If SIG114 <> RED Then SIG106 = GRN: TUMCNT106 = 0: GoTo S94
      If FS114 = EAST Then SIG106 = YEL: TUMCNT106 = 0: GoTo S94
S106RED: If SIG106 = RED Then GoTo S94
      If TUMCNT106 = TUMMAX Then SIG106 = RED: TUMCNT106 = 0: GoTo S94
      TUMCNT106 = TUMCNT106 + 1

Rem**Calculate Signal 94 Head Block Permissive
S94:  If BK12 = OCC Then GoTo S94RED
      If (TU1 = TUR Or TU2 = TUR) Then GoTo S94RED
      If SIG106 <> RED Then SIG94 = GRN Else SIG94 = YEL
      TUMCNT94 = 0: GoTo S88
S94RED: If SIG94 = RED Then GoTo S88
      If TUMCNT94 = TUMMAX Then SIG94 = RED: TUMCNT94 = 0: GoTo S88
      TUMCNT94 = TUMCNT94 + 1

Rem**Calculate Signal 88 Approach to Head Block
S88:  If FS87 = WEST Then GoTo S88RED
      If BK11 = OCC Then GoTo S88RED
      If SIG94 <> RED Then SIG88 = GRN Else SIG88 = YEL
      TUMCNT88 = 0: GoTo S76
S88RED: If SIG88 = RED Then GoTo S76
      If TUMCNT88 = TUMMAX Then SIG88 = RED: TUMCNT88 = 0: GoTo S76
      TUMCNT88 = TUMCNT88 + 1

Rem**Calculate Signal 76 Permissive in open country
S76:  If FS179 = WEST Then GoTo S76RED
      If BK10 = OCC Then GoTo S76RED
      If SIG88 <> RED Then SIG76 = GRN: TUMCNT76 = 0: GoTo S87
      If FS88 = EAST Then SIG76 = YEL: TUMCNT76 = 0: GoTo S87
S76RED: If SIG76 = RED Then GoTo S87
      If TUMCNT76 = TUMMAX Then SIG76 = RED: TUMCNT76 = 0: GoTo S87
      TUMCNT76 = TUMCNT76 + 1

Rem**CALCULATE WESTBOUND SIGNAL ASPECTS**
Rem**Calculate Signal 87 Permissive in open country
S87:  	  
--]=]
--
-- In the next 4 functions:
-- Only ThisSig is changed
-- OppositeSig and NextSig are never changed
--
function PermissiveinOpenCountry(NextBlock, ThisSig, OppositeSig, NextSig, MyDir, OppositeDir)	-- Uses Traffic Sticks
	-- This function never changes OppositeSig
	-- this function never changes NextSig, NextSig can be zero
	--print("PermissiveinOpenCountry(" .. NextBlock .. "," .. ThisSig .. "," .. OppositeSig .. "," .. NextSig .. "," .. (MyDir or "nil") .. "," .. OppositeDir .. ")");

	if (FSR[OppositeSig] == nil) then Stop("Traffic Stick nil in PermissiveinOpenCountry - Opposite Side #" .. OppositeSig); end;
	if (NextSig ~= 0 and FSR[NextSig] == nil) then Stop("Traffic Stick nil in PermissiveinOpenCountry  - Next Signal This Side # " .. NextSig); end;
	----------------------------------------------------------------------------
	-- Calculate Signal Permissive in open country using Traffic stick on opposite and next signal    
	if (FSR[OppositeSig] == OppositeDir) then goto POCAsStopandProceed; end;
    if (BlockOccupancy[NextBlock] ~= UNOCCUPIED) then goto POCAsStopandProceed; end;
	-- Check for both kinds of Stop Aspects, if there is no NextSig (== 0) then set ThisSig to AsClear
    if (Sig[NextSig] ~= AsStopSignal and Sig[NextSig] ~= AsStopandProceed) then Sig[ThisSig] = AsClear; TUMCNT[ThisSig] = 0; goto POCDone; end;
    if (FSR[NextSig] == MyDir) then Sig[ThisSig] = AsApproach; TUMCNT[ThisSig] = 0; goto POCDone; end;	-- if NextSig == 0, never gets to this line
::POCAsStopandProceed::
    if (Sig[ThisSig] == AsStopandProceed) then goto POCDone; end;
    if (TUMCNT[ThisSig] == TUMMAX) then Sig[ThisSig] = AsStopandProceed; TUMCNT[ThisSig] = 0; goto POCDone; end;
    TUMCNT[ThisSig] = TUMCNT[ThisSig] + 1;
::POCDone::	  
	  return;
end;

function HeadBlockPermissive(NextBlock, ThisSig, OppositeSig, NextSig, Switch1, Switch2, MyDir, OppositeDir)
	-- This function never accesses or changes OppositeSig
	-- this function never changes NextSig	
	--print("HeadBlockPermissive(" .. NextBlock .. "," .. ThisSig .. "," .. (OppositeSig or "nil") .. "," .. NextSig .. "," .. (MyDir or "nil") .. "," .. (OppositeDir or "nil") .. ")");
	--Calculate Signal Head Block Permissive - no Traffic Stick involved
	--print("HBP Next Block ", BlockName[NextBlock], " BlockOccupancy[NextBlock] ", BlockOccupancy[NextBlock]);
	----------------------------------------------------------------------------
    if (BlockOccupancy[NextBlock] ~= UNOCCUPIED) then goto HBPAsStopandProceed; end;
    --if (GetSwitchStatus(Switches[Switch1][CHAN_NUMBER], Switches[Switch1][AIU_NUMBER], Switches[Switch1][TIU_NUMBER]) == REVERSE or 
	--	GetSwitchStatus(Switches[Switch2][CHAN_NUMBER], Switches[Switch2][AIU_NUMBER], Switches[Switch2][TIU_NUMBER]) == REVERSE) then
	--	goto HBPAsStopandProceed;
	--  end;
    if (GetSwitchStatus(Switches[Switch1][CHAN_NUMBER], Switches[Switch1][AIU_NUMBER], Switches[Switch1][TIU_NUMBER]) == REVERSE) then
		goto HBPAsStopandProceed; end;
	-- Check for both kinds of Stop Aspects		
    if (Sig[NextSig] ~= AsStopSignal and Sig[NextSig] ~= AsStopandProceed) then Sig[ThisSig] = AsClear;
		else Sig[ThisSig] = AsApproach; end
    TUMCNT[ThisSig] = 0;
	goto HBPDone;	  
::HBPAsStopandProceed::
    if (Sig[ThisSig] == AsStopandProceed) then goto HBPDone; end;
    if (TUMCNT[ThisSig] == TUMMAX) then Sig[ThisSig] = AsStopandProceed; TUMCNT[ThisSig] = 0; goto HBPDone; end;
    TUMCNT[ThisSig] = TUMCNT[ThisSig] + 1;
::HBPDone::
	return;
end;

function HeadBlockAbsolute(NextBlock, ThisSig, OppositeSig, NextSig, MyDir, OppositeDir)	-- Uses Traffic Sticks
	-- This function never accesses or changes OppositeSig
	-- this function never changes NextSig
	--print("HeadBlockAbsolute(" .. NextBlock .. "," .. ThisSig .. "," .. (OppositeSig or "nil") .. "," .. NextSig .. "," .. (MyDir or "nil") .. "," .. (OppositeDir or "nil") .. ")");
	-- THIS FUNCTION FAILS if less than 2 blocks between sidings
	if (FSR[NextSig] == nil) then
		Stop("Traffic Stick nil in HeadBlockAbsolute - Next Signal This Side #" .. NextSig);
		end;
	----------------------------------------------------------------------------
	--Calculate Signal Head Block Absolute using Traffic Stick on next signal
    if (BlockOccupancy[NextBlock] ~= UNOCCUPIED) then goto HBAAsStop; end; -- Stop - next block is occupied by virtue of a train being in the block
	-- Otherwise the next block is unoccupied
	--
	-- if next signal (signal in advance of ThisSig) is not Stop, then set signal to clear indicating we are clear two blocks ahead with no opposing traffic
	-- Check for both kinds of Stop Aspects
    if (Sig[NextSig] ~= AsStopSignal and Sig[NextSig] ~= AsStopandProceed) then Sig[ThisSig] = AsClear; TUMCNT[ThisSig] = 0; goto HBADone; end;
	-- if traffic stick ahead indicates I'm following traffic then set signal to Approach indicating we are clear one block ahead and must prepare to stop.
    if (FSR[NextSig] == MyDir) then Sig[ThisSig] = AsApproach; TUMCNT[ThisSig] = 0; goto HBADone; end;
	-- if next signal is Stop and traffic stick indicates opposing traffic, then set signal to Stop by falling through to next code
::HBAAsStop::
    if (Sig[ThisSig] == AsStopSignal) then goto HBADone; end;
    if (TUMCNT[ThisSig] == TUMMAX) then Sig[ThisSig] = AsStopSignal; TUMCNT[ThisSig] = 0; goto HBADone; end;
    TUMCNT[ThisSig] = TUMCNT[ThisSig] + 1;
::HBADone::
	return;
end;

function ApproachtoHeadBlock(NextBlock, ThisSig, OppositeSig, NextSig, MyDir, OppositeDir)	-- Uses Traffic Sticks
	-- This function never changes OppositeSig
	-- this function never changes NextSig, NextSig can be zero
	--print("ApproachtoHeadBlock(" .. NextBlock .. "," .. ThisSig .. "," .. OppositeSig .. "," .. NextSig .. "," .. (MyDir or "nil") .. "," .. OppositeDir .. ")");
	if (FSR[OppositeSig] == nil) then Stop("Traffic Stick nil in ApproachtoHeadBlock - Opposite Side #" .. OppositeSig); end;
	----------------------------------------------------------------------------
	--Calculate Signal Approach to Head Block using Traffic Stick on opposite signal
    if (FSR[OppositeSig] == OppositeDir) then goto AHBAsStopandProceed; end;
    if (BlockOccupancy[NextBlock] ~= UNOCCUPIED) then goto AHBAsStopandProceed; end;
	-- Check for both kinds of Stop Aspects
    if (NextSig == 0 or (Sig[NextSig] ~= AsStopSignal and Sig[NextSig] ~= AsStopandProceed)) then Sig[ThisSig] = AsClear; else Sig[ThisSig] = AsApproach; end
    TUMCNT[ThisSig] = 0; 
	goto AHBDone;
::AHBAsStopandProceed::
    if (Sig[ThisSig] == AsStopandProceed) then goto AHBDone; end;
    if (TUMCNT[ThisSig] == TUMMAX) then Sig[ThisSig] = AsStopandProceed; TUMCNT[ThisSig] = 0; goto AHBDone; end;
    TUMCNT[ThisSig] = TUMCNT[ThisSig] + 1;
::AHBDone::	  
	return;
end;

--[[---------------------------------------------------------------------------------------]]

function loop(Engine, TIU)
-- Always call Sleep(X) at least once in this function if it returns true;
--print("loop() Running ", RunTime(), " seconds");

-- Read Occupancy flags
-- This is all done in the tag() function starting from initial occupancy conditions

-- SET TRAFFIC STICK UPON SEQUENTIAL TRAIN MOVEMENT
-- parameters :  SetTrafficStick(west side block, east side block, WB signal #, EB signal #)
--
if (Layout4.LLLoop == true) then
	SetTrafficStick(BALLOON_BLOCK, YOUNGSTOWN_BLOCK, 1, 2);	-- Loop Configuration
	else
	SetTrafficStick(GATEWAY_YARD_BLOCK, YOUNGSTOWN_BLOCK, 1, 2);	-- Point-to-Point Configuration
	end
-- 2 blocks between sidings
SetTrafficStick(LOWELLVILLE_BLOCK, NEW_CASTLE_JUNCTION_BLOCK, 11, 12);
SetTrafficStick(BEAVER_FALLS_BLOCK, FALLSTON_BLOCK, 19, 20);
SetTrafficStick(FALLSTON_BLOCK, MONACA_BLOCK, 21, 22);
SetTrafficStick(MONACA_BLOCK, WEST_ALIQUIPPA_BLOCK, 25, 26);
SetTrafficStick(WEST_ALIQUIPPA_BLOCK, ALIQUIPPA_BLOCK, 27, 28);
if (Layout4.LLLoop == true) then
	SetTrafficStick(MCKEES_ROCKS_BLOCK, BALLOON_BLOCK, 37, 38);	-- Loop Configuration
	else
	SetTrafficStick(MCKEES_ROCKS_BLOCK, MCKEES_ROCKS_REPAIR_SHOPS_YARD_BLOCK, 37, 38);	-- Point-to-Point Configuration
	end

-- CALCULATE WESTBOUND SIGNAL ASPECTS
if (Layout4.LLLoop == true) then
	ApproachtoHeadBlock(MCKEES_ROCKS_BLOCK, 37, 38, 33, nil, EAST);	-- Loop Configuration
	else
	ApproachtoHeadBlock(MCKEES_ROCKS_BLOCK, 37, 38, 33, nil, EAST);	-- Point-to-Point Configuration
	end;
HeadBlockPermissive(J_L_BLOCK, 33, nil, 29, ALIQUIPPA_WEST, ALIQUIPPA_EAST, nil, nil);
HeadBlockAbsolute(ALIQUIPPA_BLOCK, 29, nil, 27, WEST, nil);
PermissiveinOpenCountry(WEST_ALIQUIPPA_BLOCK, 27, 28, 25, WEST, EAST);
PermissiveinOpenCountry(MONACA_BLOCK, 25, 26, 21, WEST, EAST); 
PermissiveinOpenCountry(FALLSTON_BLOCK, 21, 22, 19, WEST, EAST);
ApproachtoHeadBlock(BEAVER_FALLS_BLOCK, 19, 20, 17, nil, EAST);
HeadBlockPermissive(COLLEGE_BLOCK, 17, nil, 13, COLLEGE_WEST, COLLEGE_EAST, nil, nil);
HeadBlockAbsolute(NEW_CASTLE_JUNCTION_BLOCK, 13, nil, 11, WEST, nil);
ApproachtoHeadBlock(LOWELLVILLE_BLOCK, 11, 12, 9, nil, EAST);
HeadBlockPermissive(STRUTHERS_BLOCK, 9, nil, 5, STRUTHERS_WEST, STRUTHERS_EAST, nil, nil);
-- does not change NextSig #1
HeadBlockAbsolute(YOUNGSTOWN_BLOCK, 5, nil, 1, WEST, nil);
if (Layout4.LLLoop == true) then
	PermissiveinOpenCountry(BALLOON_BLOCK, 1, 2, 37, WEST, EAST);	-- Loop Configuration
	else
	PermissiveinOpenCountry(GATEWAY_YARD_BLOCK, 1, 2, 0, WEST, EAST);	-- Point-to-Point Configuration
	end

-- CALCULATE EASTBOUND SIGNAL ASPECTS
if (Layout4.LLLoop == true) then
	-- does not change OppositeSig #1
	ApproachtoHeadBlock(YOUNGSTOWN_BLOCK, 2, 1, 6, nil, WEST);	-- Loop Configuration
	else
	ApproachtoHeadBlock(YOUNGSTOWN_BLOCK, 2, 1, 6, nil, WEST);	-- Point-to-Point Configuration
	end;
HeadBlockPermissive(STRUTHERS_BLOCK, 6, nil, 10, STRUTHERS_EAST, STRUTHERS_WEST, nil, nil);
HeadBlockAbsolute(LOWELLVILLE_BLOCK, 10, nil, 12, EAST, nil);
ApproachtoHeadBlock(NEW_CASTLE_JUNCTION_BLOCK, 12, 11, 14, nil, WEST);
HeadBlockPermissive(COLLEGE_BLOCK, 14, nil, 18, COLLEGE_EAST, COLLEGE_WEST,	nil, nil);
HeadBlockAbsolute(BEAVER_FALLS_BLOCK, 18, nil, 20, EAST, nil);
PermissiveinOpenCountry(FALLSTON_BLOCK, 20, 19, 22, EAST, WEST);
PermissiveinOpenCountry(MONACA_BLOCK, 22, 21, 26, EAST, WEST);
PermissiveinOpenCountry(WEST_ALIQUIPPA_BLOCK, 26, 25, 28, EAST, WEST);
ApproachtoHeadBlock(ALIQUIPPA_BLOCK, 28, 27, 30, nil, WEST);
HeadBlockPermissive(J_L_BLOCK, 30, nil, 34, ALIQUIPPA_EAST, ALIQUIPPA_WEST, nil, nil);
HeadBlockAbsolute(MCKEES_ROCKS_BLOCK, 34, nil, 38, EAST, nil);
if (Layout4.LLLoop == true) then
	PermissiveinOpenCountry(BALLOON_BLOCK, 38, 37, 2, EAST, WEST);	-- Loop Configuration
	else
	PermissiveinOpenCountry(MCKEES_ROCKS_REPAIR_SHOPS_YARD_BLOCK, 38, 37, 0, EAST, WEST);	-- Point-to-Point Configuration
	end

-- CLEAR TRAFFIC STICK FOLLOWING SIGNAL UPGRADES

if (Sig[1] ~= AsStopSignal and Sig[1] ~= AsStopandProceed and TUMCNT[1] == 0) then FSR[1] = NDT; end;
if (Sig[2] ~= AsStopSignal and Sig[2] ~= AsStopandProceed and TUMCNT[2] == 0) then FSR[2] = NDT; end;

if (Sig[11] ~= AsStopSignal and Sig[11] ~= AsStopandProceed and TUMCNT[11] == 0) then FSR[11] = NDT; end;
if (Sig[12] ~= AsStopSignal and Sig[12] ~= AsStopandProceed and TUMCNT[12] == 0) then FSR[12] = NDT; end;

if (Sig[19] ~= AsStopSignal and Sig[19] ~= AsStopandProceed and TUMCNT[19] == 0) then FSR[19] = NDT; end;
if (Sig[20] ~= AsStopSignal and Sig[20] ~= AsStopandProceed and TUMCNT[20] == 0) then FSR[20] = NDT; end;

if (Sig[21] ~= AsStopSignal and Sig[21] ~= AsStopandProceed and TUMCNT[21] == 0) then FSR[21] = NDT; end;
if (Sig[22] ~= AsStopSignal and Sig[22] ~= AsStopandProceed and TUMCNT[22] == 0) then FSR[22] = NDT; end;
	
if (Sig[25] ~= AsStopSignal and Sig[25] ~= AsStopandProceed and TUMCNT[25] == 0) then FSR[25] = NDT; end;
if (Sig[26] ~= AsStopSignal and Sig[26] ~= AsStopandProceed and TUMCNT[26] == 0) then FSR[26] = NDT; end;
	
if (Sig[27] ~= AsStopSignal and Sig[27] ~= AsStopandProceed and TUMCNT[27] == 0) then FSR[27] = NDT; end;
if (Sig[28] ~= AsStopSignal and Sig[28] ~= AsStopandProceed and TUMCNT[28] == 0) then FSR[28] = NDT; end;
	
if (Sig[37] ~= AsStopSignal and Sig[37] ~= AsStopandProceed and TUMCNT[37] == 0) then FSR[37] = NDT; end;
if (Sig[38] ~= AsStopSignal and Sig[38] ~= AsStopandProceed and TUMCNT[38] == 0) then FSR[38] = NDT; end;
	
-- Display signal aspects
for Signal = 1, 38 do
	if (Sig[Signal] ~= nil) then	-- nil means no change
		if (SavedSig[Signal] ~= Sig[Signal]) then	-- if no change, don't need to SetAspect
			SetAspect(Signal, Sig[Signal]);
			--print("SetAspect on signal # ", Signal, " to ", AspectNames[Sig[Signal]]);
			print("(", RunTime(),") SetAspect on signal # ", Signal, " to ", AspectNames[Sig[Signal]], " was ", AspectNames[SavedSig[Signal]]);
			SavedSig[Signal] = Sig[Signal];	-- save current aspect
			end;
		end;
	-- TODO consider saving signal here
	end;
-- consider approach lighting but that is hard to do since
-- we don't have a connection between block and the signals at each end
-- of the block (on opposite sides)
	
Sleep(0.1);

return true;
end
--[[---------------------------------------------------------------------------------------]]

function tag(Detector, Reader, EngineNo, TagLocation, CarModel, Railroad, CarNumber, TagPacket)
-- This function is called each time a tag programmed with 32 bytes of data is detected (TAGDATABLOCK4)
-- TagPacket contains 8 characters (4 hex digits) of UID and 32 characters
-- of block 4 data from the programmed tag.

BumpCounter(COUNTER01);	
local PacketLen = #TagPacket;

if (Debug() >= 8) then print("Packet Length = ", PacketLen); end;
if (Debug() >= 7) then print(TagPacket); end;

	-- index the table Tags[] by tag ID, value is the tag packet itself
	Tags[string.sub(TagPacket, 1, 8)] = string.sub(TagPacket, 9, 40);	-- remove checksum and EOP
	TagCount = TagCount + 1;	-- count the number of tags
	if (EngineNo > 0) then	-- its an engine
		if (BeepOn) then Beep(ASTERISK); end;
		-- next line shows a different way to quote a string using brackets
		printc(clGreen, string.format([[tag(%.2f) : Detector %d Reader %d Eng#%d %s %s #%s %s]],
			RunTime(), Detector, Reader, EngineNo, GetRailroadName(Railroad), GetEngineName(CarModel), CarNumber, GetTagLocation(TagLocation)));
		-- follow on to process the detected engine
		else	-- it's a car
		if (BeepOn) then Beep(EXCLAMATION); end;
		printc(clBlue, string.format([[tag(%.2f) : Detector %d Reader %d Car %s %s #%s]],
			RunTime(), Detector, Reader, GetRailroadName(Railroad), GetCarName(CarModel), CarNumber));
		-- follow on to process non-engine tags
		end

	-- The complete packet received is in TagPacket 
	-- (all of the fields have already been extracted into the first parameters to tag()
	--   First 8 digits are the 4 byte Tag UID in hexidecimal
	--   Next 32 digits are the 16 bytes of Block 4 Information read from the Tag in hexidecimal
	if (Debug() >= 6) then print(string.format("Packet  %s  %s", string.sub(TagPacket, 1, 8),
		string.sub(TagPacket,9,40) )); end;

-- Tag 3-0
if (Detector == MCKEES_ROCKS_REPAIR_SHOPS_YARD_DETECTOR and Reader == MCKEES_ROCKS_REPAIR_SHOPS_YARD_READER) then
	if (Func30 == nil) then
		Func30 = createMainLine(false);
		end;
	if (Layout4.LLLoop == true) then
		-- Loop Configuration
		Func30("@ MCKEES_ROCKS_REPAIR_SHOPS_YARD_TAG ", EngineNo, TagLocation, CarModel, BALLOON_BLOCK, MCKEES_ROCKS_BLOCK);
		else
		-- Point-to-Point Configuration
		Func30("@ MCKEES_ROCKS_REPAIR_SHOPS_YARD_TAG ", EngineNo, TagLocation, CarModel, MCKEES_ROCKS_REPAIR_SHOPS_YARD_BLOCK, MCKEES_ROCKS_BLOCK);
		end

-- Tag 5-1
elseif (Detector == ALIQUIPPA_EAST_DETECTOR and Reader == ALIQUIPPA_EAST_READER) then
	if (Func51 == nil) then
		Func51 = createMainLine(false);
		end;
	Func51("@ ALIQUIPPA_EAST_TAG ", EngineNo, TagLocation, CarModel, MCKEES_ROCKS_BLOCK, J_L_BLOCK);

-- Tag 1-0
elseif (Detector == ALIQUIPPA_WEST_DETECTOR and Reader == ALIQUIPPA_WEST_READER) then
	if (Func10 == nil) then
		Func10 = createMainLine(false); 
		end;	
	Func10("@ ALIQUIPPA_WEST_TAG ", EngineNo, TagLocation, CarModel, J_L_BLOCK, ALIQUIPPA_BLOCK); 

-- Tag 10-0
elseif (Detector == WEST_ALIQUIPPA_DETECTOR and Reader == WEST_ALIQUIPPA_READER) then
	if (Func100 == nil) then
		Func100 = createMainLine(false);
		end;
	Func100("@ WEST_ALIQUIPPA_TAG ", EngineNo, TagLocation, CarModel, ALIQUIPPA_BLOCK, WEST_ALIQUIPPA_BLOCK);

-- Tag 6-0
elseif (Detector == MONACA_EAST_DETECTOR and Reader == MONACA_EAST_READER) then
	if (Func60 == nil) then
		Func60 = createMainLine(false);
		end;
	Func60("@ MONACA_EAST_TAG ", EngineNo, TagLocation, CarModel, WEST_ALIQUIPPA_BLOCK, MONACA_BLOCK);
	
-- Tag 6-1
elseif (Detector == MONACA_WEST_DETECTOR and Reader == MONACA_WEST_READER) then
	if (Func61 == nil) then
		Func61 = createMainLine(false);	
		end;
	Func61("@ MONACA_WEST_TAG ", EngineNo, TagLocation, CarModel, MONACA_BLOCK, FALLSTON_BLOCK);

-- Tag 10-1
elseif  (Detector == FALLSTON_DETECTOR and Reader == FALLSTON_READER) then
	if (Func101 == nil) then
		Func101 = createMainLine(false);
		end;
	Func101("@ FALLSTON_TAG ", EngineNo, TagLocation, CarModel, FALLSTON_BLOCK, BEAVER_FALLS_BLOCK);
	
-- Tag 4-1
elseif (Detector == COLLEGE_EAST_DETECTOR and Reader == COLLEGE_EAST_READER) then
	if (Func41 == nil) then
		Func41 = createMainLine(false);
		end;
	Func41("@ COLLEGE_EAST_TAG ", EngineNo, TagLocation, CarModel, BEAVER_FALLS_BLOCK, COLLEGE_BLOCK);

-- Tag 2-1
elseif (Detector == COLLEGE_WEST_DETECTOR and Reader == COLLEGE_WEST_READER) then
	if (Func21 == nil) then
		Func21 = createMainLine(false);
		end;	
	Func21("@ COLLEGE_WEST_TAG ", EngineNo, TagLocation, CarModel, COLLEGE_BLOCK, NEW_CASTLE_JUNCTION_BLOCK);

-- Tag 9-0
elseif (Detector == LOWELLVILLE_DETECTOR and Reader == LOWELLVILLE_READER) then
	if (Func90 == nil) then
		Func90 = createMainLine(false);
		end;
	Func90("@ LOWELLVILLE_TAG ", EngineNo, TagLocation, CarModel, NEW_CASTLE_JUNCTION_BLOCK, LOWELLVILLE_BLOCK);

-- Tag 2-0
elseif (Detector == STRUTHERS_EAST_DETECTOR and Reader == STRUTHERS_EAST_READER) then
	if (Func20 == nil) then
		Func20 = createMainLine(false);
		end;
	Func20("@ STRUTHERS_EAST_TAG ", EngineNo, TagLocation, CarModel, LOWELLVILLE_BLOCK, STRUTHERS_BLOCK);

-- Tag 4-0
elseif (Detector == STRUTHERS_WEST_DETECTOR and Reader == STRUTHERS_WEST_READER) then
	if (Func40 == nil) then
		Func40 = createMainLine(false);
		end;
	Func40("@ STRUTHERS_WEST_TAG ", EngineNo, TagLocation, CarModel, STRUTHERS_BLOCK, YOUNGSTOWN_BLOCK);

-- Tag 8-1
elseif (Detector == GATEWAY_YARD_DETECTOR and Reader == GATEWAY_YARD_READER) then
	if (Func50 == nil) then
		Func50 = createMainLine(false);
		end;
	if (Layout4.LLLoop == true) then
		-- Loop Configuration
		Func50("@ GATEWAY_YARD_TAG ", EngineNo, TagLocation, CarModel, YOUNGSTOWN_BLOCK, BALLOON_BLOCK);
		else
		-- Point-to-Point Configuration
		Func50("@ GATEWAY_YARD_TAG ", EngineNo, TagLocation, CarModel, YOUNGSTOWN_BLOCK, GATEWAY_YARD_BLOCK);
		end
--	
	end;

return true;	-- true=continue to process tags, false=stop processing tags
end			
--[[---------------------------------------------------------------------------------------]]

function createMainLine(flag)
	local WaitforCaboose = flag;
	local MyLastBlockNo;
	local function MainLine(DetectorID, EngineNo, TagLocation, CarModel, Block1, Block2)
		---- uses globals: BlockOccupancy[]
		print("MainLine " .. BlockName[Block1] .. " (" .. Block1 .. ") " .. BlockOccupancy[Block1] ..
			" - " .. BlockName[Block2] .. " (" .. Block2 .. ") " .. BlockOccupancy[Block2] ..
			"  Detector " .. DetectorID .. " Engine " .. EngineNo);
		if (isEngine(EngineNo) and ((TagLocation == TAG_ON_FRONT_TRUCK) or
			(not WaitforCaboose and (TagLocation == TAG_ON_REAR_TRUCK)))) then		-- if its an engine
			
--[[--- Engine Detected -----]]
			if (WaitforCaboose) then
				OccButton();
				return EmergencyStop("Expected a caboose but got an engine");
				end;				
			--print("Detector " .. DetectorID .. " Engine " .. EngineNo);
			MyLastBlockNo = nil;	-- reset last block number
			if (BlockOccupancy[Block1] == UNOCCUPIED) then
				if ((BlockOccupancy[Block1] ~= UNOCCUPIED) and (BlockOccupancy[Block1] ~= 1)) then	-- At this point, occupancy can only be 0 or 1
					OccButton();
					return EmergencyStop("1 Occupancy Count is " .. BlockOccupancy[Block1] .. " in " .. BlockName[Block1] .. "(" .. Block1 .. ")");
					end;			
				BlockOccupancy[Block1] = BlockOccupancy[Block1] + 1;
				
				print("1 Occupy " .. BlockName[Block1] .. " (" .. Block1 .. ") " .. BlockOccupancy[Block1]);
				MyLastBlockNo = Block2;
				WaitforCaboose = true;	-- Save current state for the end of train detection call of this function
				return true;
				end;
			if (BlockOccupancy[Block2] == UNOCCUPIED) then
				if ((BlockOccupancy[Block2] ~= UNOCCUPIED) and (BlockOccupancy[Block2] ~= 1)) then	-- At this point, occupancy can only be 0 or 1
					OccButton();
					return EmergencyStop("2 Occupancy Count is " .. BlockOccupancy[Block2] .. " in " .. BlockName[Block2] .. "(" .. Block2 .. ")");
					end;			
				BlockOccupancy[Block2] = BlockOccupancy[Block2] + 1;
				
				print("2 Occupy " .. BlockName[Block2] .. " (" .. Block2 .. ") " .. BlockOccupancy[Block2]);
				MyLastBlockNo = Block1;
				WaitforCaboose = true;	-- Save current state for the end of train detection call of this function
				return true;
				end;
			---[[
			if (BlockOccupancy[Block1] ~= UNOCCUPIED and BlockOccupancy[Block2] ~= UNOCCUPIED) then 
				-- Special case of engine entering a siding block with the other track occupied by another train
				-- Occupy the siding and clear the head block when a end of train comes along
				if (Block1 == ALIQUIPPA_BLOCK or
					Block1 == MCKEES_ROCKS_BLOCK or
					Block1 == BEAVER_FALLS_BLOCK or
					Block1 == NEW_CASTLE_JUNCTION_BLOCK or
					Block1 == LOWELLVILLE_BLOCK or
					Block1 == YOUNGSTOWN_BLOCK) then
					if ((BlockOccupancy[Block2] ~= UNOCCUPIED) and (BlockOccupancy[Block2] ~= 1)) then	-- At this point, occupancy can only be 0 or 1
						OccButton();
						return EmergencyStop("3 Occupancy Count is " .. BlockOccupancy[Block2] .. " in " .. BlockName[Block2] .. "(" .. Block2 .. ")");
						end;					
					BlockOccupancy[Block2] = BlockOccupancy[Block2] + 1;	-- bump the siding occupancy count											

					print("(Siding) : 3 Occupy " .. BlockName[Block2] .. " (" .. Block2 .. ") " .. BlockOccupancy[Block2]);
					MyLastBlockNo = Block1;			-- Setup to decrement the head block occupancy count
					WaitforCaboose = true;			-- Save current state for the end of train detection call of this function
					return true;
					end;
				if (Block2 == ALIQUIPPA_BLOCK or
					Block2 == MCKEES_ROCKS_BLOCK or
					Block2 == BEAVER_FALLS_BLOCK or
					Block2 == NEW_CASTLE_JUNCTION_BLOCK or
					Block2 == LOWELLVILLE_BLOCK or
					Block2 == YOUNGSTOWN_BLOCK) then
					if ((BlockOccupancy[Block1] ~= UNOCCUPIED) and (BlockOccupancy[Block1] ~= 1)) then	-- At this point, occupancy can only be 0 or 1
						OccButton();
						return EmergencyStop("4 Occupancy Count is " .. BlockOccupancy[Block1] .. " in " .. BlockName[Block1] .. "(" .. Block1 .. ")");
						end;					
					BlockOccupancy[Block1] = BlockOccupancy[Block1] + 1;	-- bump the siding occupancy count	
					
					print("(Siding) : 4 Occupy " .. BlockName[Block1] .. " (" .. Block1 .. ") " .. BlockOccupancy[Block1]);
					MyLastBlockNo = Block2;			-- setup to decrement the head block occupancy count
					WaitforCaboose = true;			-- save current state for the end of train detection call of this function
					return true;
					end;
				end;
			--]]
			return EmergencyStop("MainLine : Illegal Block Status");

--[[--- Caboose Detected -----]]

		elseif (EngineNo == 0 and (CarModel == CABOOSE or CarModel == OBSERVATION) and WaitforCaboose) then
			print("Detector " .. DetectorID .. "End of Train detected");
			if ((BlockOccupancy[MyLastBlockNo] ~= 1) and (BlockOccupancy[MyLastBlockNo] ~= 2)) then	-- At this point, occupancy can only be 1 or 2
				return EmergencyStop("5 Occupancy Count is " .. BlockOccupancy[MyLastBlockNo] .. " in " .. BlockName[MyLastBlockNo] .. "(" .. MyLastBlockNo .. ")");
				end;
			BlockOccupancy[MyLastBlockNo] = BlockOccupancy[MyLastBlockNo] - 1;	-- decrement the occupancy count
			print("De-occupy " .. BlockName[MyLastBlockNo] .. " (" .. MyLastBlockNo .. ") ".. BlockOccupancy[MyLastBlockNo]);
			MyLastBlockNo = nil;	-- once MyLastBlockNo is used, set it to nil to catch possible future errors
			WaitforCaboose = false;
			writeSettings(false);	-- save state without comment
			end;
	-- other combinations of engines/cabooses/other cars are ignored
	return true;
	end;
-- return the MainLine() function when createMainLine() is called
return MainLine;
end;
--[[---------------------------------------------------------------------------------------]]

function cleanup(Engine, TIU)
	-- this function is called once when the user presses the [STOP] button or the Stop() function is called
	print("cleanup() Running");
	-- Next line not needed - This script should run independantly from Engine control scripts
	-- The line is here as it is needed when running with the test scripts Layout4 Block XX Occupancy.lua 
	ReqBlockEvents(false);	-- cancel request block occupancy flag events from the Layout
	-- save occupied block numbers
	writeSettings(true);			-- write the .ini file and save the occupied blocks
	if (ErrCount == 0) then
		print("No errors");
		else
		print(ErrCount .. " Error(s)");
		end;	
	return true;		-- false=cleanup failed, true=cleanup succeeded
	end
--[[---------------------------------------------------------------------------------------]]

function ReloadButton()
	print("Reloading layout4.lua");
	package.loaded["layout4"] = nil;	-- unload
	Layout4 = require([[layout4]]);	-- All of the details about the Layout
	return true;
end
Function11Name, Function11Label = ReloadButton, "Reload";
--[[---------------------------------------------------------------------------------------]]

function BeepOffButton()
	print("Beep Off");
	BeepOn = false;
	return true;
end
Function14Name, Function14Label = BeepOffButton, "Beep Off";
--[[---------------------------------------------------------------------------------------]]

function BeepOnButton()
	print("Beep On");
	BeepOn = true;
	return true;
end
Function15Name, Function15Label = BeepOnButton, "Beep On";
--[[---------------------------------------------------------------------------------------]]

function OccButton(Eng, TIU)
printc(clGreen, "Detection Status -------------------------");
local iBlock, FColor;
for iBlock = FIRST_BLOCK, LAST_BLOCK do
	-- passing and station tracks
	if (iBlock == STRUTHERS_STATION_TRACK) then goto continue; end;	-- passing and station tracks don't have separate occupancy status
	if (iBlock == COLLEGE_STATION_TRACK) then goto continue; end;
	if (iBlock == J_L_PASSING_TRACK) then goto continue; end;
	local flag = BlockOccupancy[iBlock];
	if (flag == UNOCCUPIED) then
		FColor = clBlack;
		else
		FColor = clRed;
		end;
	if (iBlock == STRUTHERS_THRU_TRACK) then 
		printc(FColor, "Struthers Thru & Station Tracks" .. " (" .. iBlock .. ") = ", flag);;
		goto continue;
		end;
	if (iBlock == COLLEGE_THRU_TRACK) then 
		printc(FColor, "College Thru & Station Tracks" .. " (" .. iBlock .. ") = ", flag);;
		goto continue;
		end;
	if (iBlock == J_L_THRU_TRACK) then 
		printc(FColor, "J&L Thru & Passing Tracks" .. " (" .. iBlock .. ") = ", flag);;
		goto continue;
		end;
	printc(FColor, BlockName[iBlock] .. " (" .. iBlock .. ") = ", flag);;
	::continue::
	end;	
return true;
end
Function01Name, Function01Label = OccButton, "Detection Status";
--[[---------------------------------------------------------------------------------------]]

function F02()
local response = InputBox(title, "Occupy Block Number?", ALIQUIPPA_BLOCK);
local Enum = tonumber(response);
if (Enum ~= nil and Enum ~= 0) then	-- if the result is a number, then use it, otherwise, ignore it
	if (Enum == STRUTHERS_STATION_TRACK) then Enum = STRUTHERS_BLOCK; end;	-- correct for passing and station tracks
	if (Enum == COLLEGE_STATION_TRACK) then Enum = COLLEGE_BLOCK; end;
	if (Enum == J_L_PASSING_TRACK) then Enum = J_L_BLOCK; end;
	BlockOccupancy[Enum] = BlockOccupancy[Enum] + 1;			-- This flag is used by this APB Master thread
	OverrideOccupancy(Enum, SOFTOCCUPY);						-- This flag is used by the EOB PC threads
	print("Occupy " .. BlockName[Enum] .. " (" .. Enum .. ")");
	end
return true;
end
Function02Name, Function02Label = F02, "Occupy Block";
--[[---------------------------------------------------------------------------------------]]

function F03()
local response = InputBox(title, "De-occupy Block Number?", ALIQUIPPA_BLOCK);
local Enum = tonumber(response);
if (Enum ~= nil and Enum ~= 0) then	-- if the result is a number, then use it, otherwise, ignore it
	if (Enum == STRUTHERS_STATION_TRACK) then Enum = STRUTHERS_BLOCK; end;	-- correct for passing and station tracks
	if (Enum == COLLEGE_STATION_TRACK) then Enum = COLLEGE_BLOCK; end;
	if (Enum == J_L_PASSING_TRACK) then Enum = J_L_BLOCK; end;	
	BlockOccupancy[Enum] = UNOCCUPIED;					-- this flag is used by this APB Master thread
	OverrideOccupancy(Enum, UNOCCUPIED);		-- This flag is used by the Engine on Board PC threads
	print("De-occupy " .. BlockName[Enum] .. " (" .. Enum .. ")");
	end
return true;
end
Function03Name, Function03Label = F03, "De-occupy Block";
--[[---------------------------------------------------------------------------------------]]

function F04()
for Block = FIRST_BLOCK, LAST_BLOCK do
	BlockOccupancy[Block] = UNOCCUPIED;			-- this flag is used by this APB Master thread
	OverrideOccupancy(Block, UNOCCUPIED);		-- This flag is used by the EOB PC threads
	end;
print("All blocks manually De-occupied");
return true;
end
Function04Name, Function04Label = F04, "De-occupy All Blocks";
--[[---------------------------------------------------------------------------------------]]
function LLLButton()
	if (Layout4.LLLoop == true) then
		print("Balloon Loop");
		Switch(REVERSE, Switches[BALLOON][CHAN_NUMBER], Switches[BALLOON][AIU_NUMBER], Switches[BALLOON][TIU_NUMBER], 0);	
		else
		print("Point-to-Point");
		Switch(NORMAL, Switches[BALLOON][CHAN_NUMBER], Switches[BALLOON][AIU_NUMBER], Switches[BALLOON][TIU_NUMBER], 0);	
		end;
	return true;
end
Function06Name, Function06Label = LLLButton, "Balloon/Pt-to-Pt";
--[[---------------------------------------------------------------------------------------]]
function RunTimeButton(Eng, TIU)
-- this function does not do anything except print the run time
print(string.format("RunTime = %.2f seconds", RunTime()));
return true;
end
Function07Name, Function07Label = RunTimeButton, "Run Time";
--[[---------------------------------------------------------------------------------------]]
--
-- Routine
--	1. to accept TAG_EVENT from "Layout4 Block XX Occupancy.lua"
--	2. to accept BLOCK_OCCUPANCY_EVENT from "Layout4 Block XX Occupancy.lua"
--  3. to accept BLOCK_OCCUPANCY_EVENT from manual setting of occupancy in Layout Window (Layout.cpp)
-- examples:
--	res = SetEvent(1, BLOCK_OCCUPANCY_EVENT, FLAG_NUMBER, FLAG_VALUE, 0);	-- set block occupancy flag
--	res = SetEvent(1, TAG_EVENT, DetectorNo, ReaderNo, ENGINE_OR_CABOOSE);	-- simulate a tag
--
function event(FromPC, Type, P1, P2, P3)

if (Type == BLOCK_OCCUPANCY_EVENT) then
	-- Block Number must be non zero and valid flag number
	if (P1 < 1 or P1 > FLAGLIMIT) then
		return false;
		end;
	if (FromPC > PCMAX) then
		-- FromPC == 99 for layout window
		print("Block Occupancy event() From Layout Window - ", BlockName[P1], " Value = ", P2);
		else
		-- FromPC 1 to PCMAX for PC Window
		print("Block Occupancy event() From PC# ", FromPC, " - ", BlockName[P1], " Value = ", P2);
		end
	-- Block Number = P1, value can be 1 to (FLAGLIMIT - 1)
	-- Flag Value = P2, value can be -2, -1, 0 , 1, 2
	-- Flag values of -2 or -1 are changed to a 1
	-- P3 not used
	-- update the block number for passing and station tracks
	if (P1 == STRUTHERS_STATION_TRACK) then 
		P1 = STRUTHERS_BLOCK; 
		elseif (P1 == COLLEGE_STATION_TRACK) then 
			P1 = COLLEGE_BLOCK; 
			elseif (P1 == J_L_PASSING_TRACK) then 
				P1 = J_L_BLOCK; 
				end;

	-- An event can set any occupancy flag to any value and if value equals 1 then increment block occupancy count
	if (P2 <= 0) then
		BlockOccupancy[P1] = P2;
		else
		BlockOccupancy[P1] = BlockOccupancy[P1] + P2;
		end
	-- It is the responsibility of the sketch or program sending the event to call SetFlag()
	-- in order to set the flag to the correct color and to set the hint correctly.
	end;
if (Type == TAG_EVENT) then
	-- Detector = P1
	-- Reader	= P2
	--            P3 ~= 0 means Engine
	--            P3 == 0 means Caboose --- NOTE that this supports only an Engine tag and a Caboose tag
	-- Engine Number	1561				NOT_AN_ENGINE
	-- CarModel			UNKNOWN_CARMODEL	CABOOSE
	-- Car Number		1561				123
	if (P3 ~= NOT_AN_ENGINE) then	-- engine
		print("Engine Tag Event");
		tag(P1, P2, 1561, TAG_ON_FRONT_TRUCK, UNKNOWN_CARMODEL, P_AND_LE, "1561", "NONE")	;
		else			-- caboose
		print("Caboose Tag Event");
		tag(P1, P2, NOT_AN_ENGINE, TAG_ON_FRONT_TRUCK, CABOOSE, P_AND_LE, "123", "NONE")	;
		end;
	end;
return true;		-- false=event failed, true=event succeeded	
end

--[[---------------------------------------------------------------------------------------]]
